VERSION = (
    1,
    3,
    0,
)

__title__ = "rpcm"
__description__ = "Rational Polynomial Camera Model for optical satellite images."
__url__ = 'https://github.com/cmla/rpcm'
__version__ = ".".join(map(str, VERSION))
__author__ = "Carlo de Franchis"
__author_email__ = 'carlo.de-franchis@ens-cachan.fr'
